# -*- coding: utf-8 -*-
'''Core commands for aakbar.
'''

# standard library imports
import os
import math
import csv

# external packages
import pandas as pd

# module imports
from .common import *
from . import cli, get_user_context_obj, logger

@cli.command()
@click.argument('infilestem', type=str)
@click.argument('sigset', type=str)
@click.argument('setlist', nargs=-1, type=DATA_SET_VALIDATOR)
def conserved_signature_stats(infilestem, sigset, setlist):
    """Stats on signatures found in all input genomes.

    :param: infilestem: Input filestem.
    :param sigset: Name of signature set used.
    :param setlist: List of data sets.
    :return:
    """
    global config_obj
    user_ctx = get_user_context_obj()
    if user_ctx['first_n']:
        logger.info('Only first %d records will be used.', user_ctx['first_n'])
        first_n = user_ctx['first_n']
    else:
        first_n = 1E12
    # parameter inputs
    filestem = infilestem + '-' + sigset
    infilename = filestem + '_sigcounts.tsv'
    logger.debug('Input file name is "%s".', infilename)
    setlist = DATA_SET_VALIDATOR.multiple_or_empty_set(setlist)
    #
    # Read input terms from signature set.
    #
    intersect_dir = config_obj.config_dict['summary']['dir']
    termfilepath = os.path.join(intersect_dir, sigset+'_terms.tsv')
    logger.debug('Reading terms from file "%s".', termfilepath)
    if not os.path.exists(termfilepath):
        logger.error('input file "%s" does not exist.', termfilepath)
        sys.exit(1)
    term_frame = pd.DataFrame().from_csv(termfilepath, sep='\t')
    del term_frame['score']
    n_terms = len(term_frame)
    k = len(term_frame.index[0])
    n_intersections = max(term_frame['intersections'])
    logger.info('Highly-conserved terms (HCterms) are those occur in %d genomes.', n_intersections)
    logger.info('%d %d-mer terms defined.', n_terms,
                k)
    logger.info('Calculating HCterms for %d data sets:', len(setlist))
    #
    # operate on highly-conserved terms only
    #
    hc_terms = term_frame[term_frame['intersections'] == n_intersections]
    del term_frame
    del hc_terms['intersections']
    n_hc_terms = len(hc_terms)
    logger.info('%d HCterms in signature set.', n_hc_terms)
    #
    # loop on sets
    #
    for calc_set in setlist:
        logger.info('set %s:', calc_set)
        dir = config_obj.config_dict[calc_set]['dir']
        infilepath = os.path.join(dir, infilename)
        if not os.path.exists(infilepath):
            logger.error('Input file "%s" does not exist.', infilepath)
            sys.exit(1)
        sig_frame = pd.read_csv(infilepath,
                                usecols=[0, 3],
                                index_col=0,
                                sep='\t')
        n_sigs = 0
        n_hc_sigs = 0
        sigs_in_hc = set()
        for sig in sig_frame.index:
            if sig in hc_terms.index:
                sigs_in_hc.add(sig)
                n_hc_sigs += 1
                #my_lca_counts.loc[sig] = sig_frame.loc[sig]['counts']
                #term_stats = lca_terms.loc[sig]
                #logger.info(sig_frame.loc[sig]['counts'])
            if n_sigs > first_n:
                break
            n_sigs += 1
        bins = [1,2,3,4,5,6,7,8,
                10,12,14,16,
                20,24,28,32,
                40,48,56,64,
                80,96,112,128,
                160,192,224,256]
        lastbin = 0
        #
        # Fraction found historam initialization
        #
        fractionfoundpath = os.path.join(dir, filestem + '_fractionfound.tsv')
        fractionfoundfh = open(fractionfoundpath, 'wt')
        fractionfoundwriter = csv.DictWriter(fractionfoundfh,
                                            fieldnames=['bin',
                                                        'n_found',
                                                        'found_percent',
                                                        'sigma_percent'],
                                            delimiter='\t')
        fractionfoundwriter.writeheader()
        #
        # Missing list initialization
        #
        missinglistpath = os.path.join(dir, filestem + '_missingsigs.tsv')
        missinglistfh = open(missinglistpath, 'wt')
        missinglistwriter = csv.DictWriter(missinglistfh,
                                            fieldnames=['signature',
                                                        'avg_count',
                                                        'max_count'],
                                            delimiter='\t')
        missinglistwriter.writeheader()
        #
        logger.info('   bin\tHCsigs\tfraction\t+/-')
        for i in range(len(bins)):
            nextbin = bins[i]
            inrange = hc_terms[hc_terms['max_count'].isin([lastbin,nextbin])].index
            rangeset = set([term for term in inrange])
            n_terms_in_bin = len(rangeset)
            if n_terms_in_bin == 0:
                continue
            sigs_in_range = rangeset.intersection(sigs_in_hc)
            for missing in rangeset.difference(sigs_in_hc):
                missinglistwriter.writerow({
                    'signature': missing,
                    'avg_count': hc_terms.loc[missing]['count']/float(n_intersections),
                    'max_count': hc_terms.loc[missing]['max_count']
                })
            n_sigs_in_bin = len(sigs_in_range)
            fraction_found = 100.*n_sigs_in_bin/n_terms_in_bin
            sigma = math.sqrt(n_sigs_in_bin)
            sigma_percent = 100.*sigma/n_terms_in_bin
            logger.info('   %d: %d\t%.1f\t%.1f',
                        nextbin,
                        n_sigs_in_bin,
                        fraction_found,
                        sigma_percent)
            fractionfoundwriter.writerow({
                'bin': nextbin,
                'n_found': n_sigs_in_bin,
                'found_percent': fraction_found,
                'sigma_percent': sigma_percent
            })
            lastbin = nextbin
        logger.info('   %d (%.1f%% of overall) highly-conserved sigs in %s.',
                    n_hc_sigs,
                    n_hc_sigs * 100. / n_hc_terms,
                    calc_set)
        fractionfoundfh.close()
        missinglistfh.close()







# -*- coding: utf-8 -*-
'''Core commands for aakbar.
'''

# standard library imports
import os
import shutil
import locale
import stat

# external packages
import pkg_resources
import numpy as np
import pandas as pd
import pyfaidx

# Matplotlib -use non-interactive backend
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

# module imports
from .common import *
from . import cli, get_user_context_obj, logger, log_elapsed_time

# global definitions
ALPHABETSIZE = 20 # number of coded amino acids
UNITNAME = 'Mbasepair'
UNITMULTIPLIER = 3.E6
AMBIGUOUS_RESIDUES = ['X', '.']
NUM_HISTOGRAM_BINS = 25
DEFAULT_MAX_SCORE = 0.3

#
# Helper functions begin here.
#
def is_unambiguous(seq):
    '''Check to see if all characters (residues) are unambiguous.

    :param seq: Input sequence, possibly containing ambiguous positions ('X')
                and masked (lower-case) positions.
    :return: True if no ambiguous are contained in seq.
    '''
    if any([char in seq for char in AMBIGUOUS_RESIDUES]):
        return False
    else:
        return True


def num_masked(seq):
    """Count the number of lower-case characters in a sequence.

        :param seq: Sequence of characters.
        :type seq: str, bytes, or other convertible sequence type.
        :return: Count of lower-case characters.
        :rtype: int
    """
    gene = to_str(seq)
    mask = []
    [mask.append(gene[i].islower()) for i in range(len(gene))]
    masked = sum(mask)
    return masked



def frequency_and_score_histograms(freqs, scores, dir, filestem):
    '''Writes frequency histograms to a khmer-compatible file.

    :param freqs: Vector of frequencies.
    :param scores: Vector of scores.
    :param filepath: Output file path.
    :return:
    '''
    # calculate frequency histogram
    binvals, freq_hist = np.unique(freqs, return_counts=True)
    max_freq = max(binvals)
    hist_filepath = os.path.join(dir, filestem+'_freqhist.csv')
    logger.debug('Writing frequency histogram to %s.', hist_filepath)
    cumulative = np.cumsum(freq_hist)
    total = np.sum(freq_hist)
    pd.DataFrame({'abundance':binvals,
                  'count':freq_hist,
                  'cumulative':cumulative,
                  'cumulative_fraction':cumulative/total},
                  columns=('abundance',
                           'count',
                           'cumulative',
                           'cumulative_fraction')).to_csv(hist_filepath,
                                                         index=False,
                                                         float_format='%.3f')
    logger.info('   Maximum term frequency is %d (%.2e per unique %s).',
                max_freq,
                max_freq*UNITMULTIPLIER/len(freqs),
                UNITNAME)

    # calculate score histogram
    (score_hist, bins) = np.histogram(scores, bins=[0.,
                                                    0.01, 0.03,
                                                    0.1, 0.3,
                                                    1.0, 1.3,
                                                    2.0,3.0,4.0,5.0,100.])
    score_hist = score_hist*100./len(scores)
    score_filepath = os.path.join(dir, filestem+'_scorehist.tsv')
    logger.debug('Writing score histogram to file "%s".', score_filepath)
    pd.Series(score_hist, index=bins[:-1]).to_csv(score_filepath, sep='\t',
                                                  float_format='%.2f')


def intersection_histogram(frame, dir, filestem, plot_type, n_sets, k):
    '''Do histograms of frequencies by intersection number.

    :param frame:
    :param dir:
    :param filestem:
    :param plot_type:
    :return:
    '''
    intersect_bins = []
    lastbin = 0
    nextbin = 1
    hists = {}
    intersect_filepath = os.path.join(dir, filestem+'_intersect.tsv')
    logger.debug('Writing intersection frequency histograms to %s.', intersect_filepath)
    max_freq = max(frame['max_count'])
    while nextbin < max_freq:
        inrange = frame[frame['max_count'].isin([lastbin,nextbin])]
        ibins, ifreqs = np.unique(inrange['intersections'], return_counts=True)
        intersect_bins.append(nextbin)
        hists[nextbin] = pd.Series(ifreqs, index=ibins)
        lastbin = nextbin
        nextbin *= 2
    intersect_frame = pd.DataFrame(hists).transpose().fillna(0).astype(int)
    if len(intersect_frame) == 0:
        logger.info('zero intersections for %s, skipping', dir)
        return
    intersect_frame.to_csv(intersect_filepath, sep='\t')
    #
    # plot intersection histograms
    #
    plot_filepath = os.path.join(dir, filestem+'_intersect.'+ plot_type)
    logger.debug('Plotting intersection histograms to %s.', plot_filepath)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    xvals = np.array(list(range(2, n_sets+1)), dtype=np.int32)
    for i in range(len(intersect_frame)):
        bin_edge = intersect_frame.index[i]
        data = intersect_frame.iloc[i]
        sum = data.sum()
        if i == 0:
            label = 'Singletons (%s)' %locale.format('%d',
                                                     sum,
                                                     grouping=True)
        else:
            label='%d+/genome (%s)' %(bin_edge,
                                      locale.format('%d',
                                                    sum,
                                                    grouping=True))
        ax.plot(xvals,
                data*100./sum,
                '-',
                label=label)
    ax.legend(loc=9)
    plt.xlim([2,n_sets])
    plt.title('%d-mer Intersections Across %d Genomes' %(k, n_sets))
    plt.xlabel('Number Intersecting')
    plt.ylabel('% of Shared 10-mers in Bin')
    plt.savefig(plot_filepath)


#
# Cli commands begin here.
#
@cli.command()
@click.option('-k', default=DEFAULT_K, show_default=True, help='Term length')
@click.argument('infilename', type=str)
@click.argument('outfilestem', type=str)
@click.argument('setlist', nargs=-1, type=DATA_SET_VALIDATOR)
@log_elapsed_time()
def calculate_peptide_terms(k, infilename, outfilestem, setlist):
    '''Write peptide terms and histograms.

    '''
    # context inputs
    user_ctx = get_user_context_obj()
    if user_ctx['first_n']:
        logger.info('Only first %d records will be used', user_ctx['first_n'])
    simplicity_obj = user_ctx['simplicity_object']
    logger.info('Simplicity function is %s.',
                simplicity_obj.desc)
    simplicity_obj.set_k(k)
    # parameter inputs
    logger.info('Term size is %d characters.', k)
    logger.info('Input file name is "%s".', infilename)
    logger.info('Output file stem is "%s".', outfilestem)
    setlist = DATA_SET_VALIDATOR.multiple_or_empty_set(setlist)
    logger.info('Calculating terms for %d data sets:', len(setlist))
    #
    # loop on sets
    #
    for calc_set in setlist:
        dir = config_obj.config_dict[calc_set]['dir']
        try:
            label = config_obj.config_dict[calc_set]['label']
        except KeyError:
            label = dir
        infilepath = os.path.join(dir, infilename)
        if not os.path.exists(infilepath):
            logger.error('Input file "%s" does not exist.', infilepath)
            sys.exit(1)
        #
        # calculate each unambiguous term and its score
        #
        term_score_list = []
        n_residues = 0
        n_raw_terms = 0
        fasta  = pyfaidx.Fasta(infilepath, mutable=True)
        if user_ctx['first_n']:
            keys = list(fasta.keys())[:user_ctx['first_n']]
        else:
            keys = fasta.keys()
        n_recs = len(keys)
        #
        # loop on genes, with or without progress bars
        #
        if user_ctx['progress']:
            with click.progressbar(keys, label='   %s genes processed'
                                   %label, length=n_recs) as bar:
                for key in bar:
                    seq = fasta[key]
                    s_scores = np.array(simplicity_obj.score(seq))
                    seq = to_str(seq).upper()
                    seqlen = len(seq)
                    n_potential_terms = seqlen - k + 1
                    if n_potential_terms < 1:
                        logger.info('   skipping gene $s of length %d',
                                    key, seqlen)
                        continue
                    n_residues += seqlen
                    n_raw_terms += n+potential_terms
                    term_score_list += [ (str(seq[i:i+k]), s_scores[i])
                                         for i in range(seqlen-k)
                                         if is_unambiguous(str(seq[i:i+k]))]
        else:
            logger.info('  %s: ', label)
            for key in keys:
                seq = fasta[key]
                s_scores = np.array(simplicity_obj.score(seq))
                seq = to_str(seq).upper()
                seqlen = len(seq)
                n_potential_terms = seqlen - k + 1
                if n_potential_terms < 1:
                    logger.info('   skipping gene %s of length %d',
                                key, seqlen)
                    continue
                n_residues += seqlen
                n_raw_terms += n_potential_terms
                term_score_list += [ (str(seq[i:i+k]), s_scores[i])
                                     for i in range(seqlen-k)
                                     if is_unambiguous(str(seq[i:i+k]))]

        fasta.close()
        term_arr = np.array([term for term, score in term_score_list],
                            dtype=np.dtype(('S%d'%(k))))
        score_arr = np.array([score for term, score in term_score_list],
                             dtype=np.int16)
        del term_score_list
        n_terms = len(term_arr)
        n_skipped = n_raw_terms - n_terms
        logger.info('   %s genes, %s residues, %s (%0.2f%%) ambiguous, and %s input terms.',
                    locale.format('%d', n_recs, grouping=True),
                    locale.format('%d', n_residues, grouping=True),
                    locale.format('%d', n_skipped, grouping=True),
                    100*n_skipped/n_raw_terms,
                    locale.format('%d', n_terms, grouping=True),
                    )
        #
        # calculate unique terms
        #
        sort_arr = np.argsort(term_arr)
        term_arr = term_arr[sort_arr]
        score_arr = score_arr[sort_arr]
        unique_terms, beginnings, freqs = np.unique(term_arr,
                                        return_index=True,
                                        return_counts=True)
        mean_scores = np.array([score_arr[beginnings[i]:beginnings[i]+freqs[i]].mean()
                                for i in range(len(beginnings))],
                               dtype=np.float32)
        n_unique = len(unique_terms)
        logger.info('   %s unique terms (%.2f%% of input, %.6f%% of %s possible %d-mers).',
                    locale.format('%d', n_unique, grouping=True),
                    100.*n_unique/n_terms,
                    100.*n_unique/(ALPHABETSIZE**k),
                    locale.format('%d', ALPHABETSIZE**k, grouping=True),
                    k)
        #
        # write terms, counts, and scores in sorted form
        #
        term_filepath = os.path.join(dir, outfilestem+'_terms.tsv')
        logger.debug('writing unique terms and counts to %s', term_filepath)
        sort_arr = np.argsort(freqs)
        pd.DataFrame({'count':freqs[sort_arr],
                      'score':mean_scores[sort_arr]},
                     index=[i.decode('UTF-8') for i in unique_terms[sort_arr]],
                     ).to_csv(term_filepath,
                              sep='\t',
                              float_format='%.2f')
        del sort_arr
        # histograms
        frequency_and_score_histograms(freqs, mean_scores, dir, outfilestem)


@cli.command()
@click.option('--cutoff', default=DEFAULT_MAX_SCORE, show_default=True,
              help='Maximum simplicity score to keep.')
@click.argument('infilestem', type=str)
@click.argument('outfilestem', type=str)
@log_elapsed_time()
def filter_peptide_terms(cutoff, infilestem, outfilestem):
    '''Removes high-simplicity terms.

    '''
    global config_obj
    dir = config_obj.config_dict['summary']['dir']
    # argument inputs
    logger.debug('Input file stem is "%s".', infilestem)
    logger.debug('Output file stem is "%s".', outfilestem)
    logger.info('Minimum simplicity value is %0.2f.', cutoff)
    #
    # Read input terms from merged set
    #
    infilepath = os.path.join(dir, infilestem+'_terms.tsv')
    logger.debug('Filtering file "%s".', infilepath)
    if not os.path.exists(infilepath):
        logger.error('input file "%s" does not exist.', infilepath)
        sys.exit(1)
    term_frame = pd.DataFrame().read_csv(infilepath, sep='\t')
    n_intersecting_terms = len(term_frame)
    k = len(term_frame.index[0])
    logger.info('   %d %d-mer terms initially.', n_intersecting_terms,
                k)
    #
    # drop terms that don't meet cutoff
    #
    term_frame.drop(term_frame[term_frame['score'] > cutoff].index,
                      inplace=True)
    n_scored_terms = len(term_frame)
    logger.info('   %s terms passing cutoff, representing',
                locale.format('%d', n_scored_terms, grouping=True))
    logger.info('       %.2f%% of %s intersecting terms, and',
                n_scored_terms * 100. / n_intersecting_terms,
                locale.format("%d", n_intersecting_terms, grouping=True))
    logger.info('       %.6f%% of possible %d-mers.',
                n_scored_terms * 100. / (ALPHABETSIZE ** k), k)
    #
    # write frequency and score histograms
    frequency_and_score_histograms(term_frame['count'],
                                   term_frame['score'],
                                   dir,
                                   outfilestem)
    #
    # write terms
    #
    term_frame.sort_values(by=['max_count', 'intersections'], inplace=True)
    term_filepath = os.path.join(dir, outfilestem + '_terms.tsv')
    logger.debug('Writing merged terms to "%s".', term_filepath)
    term_frame.to_csv(term_filepath, sep='\t',
                        float_format='%0.2f')
    #
    # calculate histogram of intersections
    #
    intersection_histogram(term_frame, dir, outfilestem,
                           config_obj.config_dict['plot_type'],
                           max(term_frame['intersections']), k)


@cli.command()
@click.argument('filestem', type=str)
@click.argument('setlist', nargs=-1, type=DATA_SET_VALIDATOR)
@log_elapsed_time()
def intersect_peptide_terms(filestem, setlist):
    '''Find intersecting terms from multiple sets.

    :param filestem: input and output filename less '_terms.tsv'
    :param setlist:
    :return:
    '''
    #
    # get configuration inputs
    #
    global config_obj
    outdir = config_obj.config_dict['summary']['dir']
    if not os.path.exists(outdir) and not os.path.isdir(outdir):
        logger.info('Making summary directory %s', outdir)
        os.makedirs(outdir)
    #
    # get argument inputs
    #
    infilename = filestem+'_terms.tsv'
    logger.info('Input filenames will be "%s".', infilename)
    setlist = DATA_SET_VALIDATOR.multiple_or_empty_set(setlist)
    n_sets = len(setlist)
    logger.info('Joining terms from %d sets:', n_sets)
    #
    # read and merge the term_lists
    #
    merged_frame = None
    n_terms_total = 0
    for calc_set in setlist:
        dir = config_obj.config_dict[calc_set]['dir']
        infilepath = os.path.join(dir, infilename)
        if not os.path.exists(infilepath):
            logger.error('input file "%s" does not exist.', infilepath)
            sys.exit(1)
        working_frame = pd.DataFrame.read_csv(infilepath,
                                              sep='\t',
                                              index_col=0)
        n_terms_set = len(working_frame)
        n_terms_total += n_terms_set
        if merged_frame is None:    # first one read
            merged_frame = pd.DataFrame({'intersections': [1]*n_terms_set,
                                         'count': working_frame['count'],
                                         'max_count': working_frame['count'],
                                         'score': working_frame['score']*working_frame['count']},
                                        index=working_frame.index,
                                        columns=('intersections', 'count', 'max_count', 'score')
                                        )
        else: # join this frame
            working_frame.rename(columns={'count':'working_count',
                                          'score':'working_score'},
                                 inplace=True)
            merged_frame = merged_frame.join(working_frame, how='outer')
            del working_frame
            merged_frame = merged_frame.fillna(0)
            merged_frame['count'] += merged_frame['working_count']
            merged_frame['max_count'] = merged_frame[['max_count','working_count']].max(axis=1)
            merged_frame['score'] += merged_frame['working_score']* merged_frame['working_count']
            merged_frame['intersections'] += (merged_frame['working_count'] > 0).astype(np.int)
            del merged_frame['working_count']
            del merged_frame['working_score']
        n_unique_terms = len(merged_frame)
        logger.info('   %s: %s terms in (%.0f%% of %s unique, %.0f%% of %s total read)',
                    calc_set,
                    locale.format("%d", n_terms_set, grouping=True),
                    100.*n_terms_set/n_unique_terms,
                    locale.format("%d", n_unique_terms, grouping=True),
                    100.*n_terms_set/n_terms_total,
                    locale.format('%d', n_terms_total, grouping=True))
    k = len(merged_frame.index[0])
    logger.info('%s unique %d-mers (%0.1f%% of %s total in).',
                locale.format("%d", n_unique_terms, grouping=True),
                k,
                100.*n_unique_terms/n_terms_total,
                locale.format('%d', n_terms_total, grouping=True))
    #
    # drop terms that don't intersect in two sets
    #
    merged_frame.drop(merged_frame[merged_frame['intersections'] == 1].index,
                      inplace=True)
    n_intersecting_terms = len(merged_frame)
    logger.info('%s intersecting terms (%.1f%% of unique).',
                locale.format('%d', n_intersecting_terms, grouping=True),
                100.*n_intersecting_terms/n_unique_terms)
    #
    # clean up and normazlize
    #
    merged_frame['count'] = merged_frame['count'].astype(np.int32)
    merged_frame['max_count'] = merged_frame['max_count'].astype(np.int32)
    merged_frame['score'] = merged_frame['score']/merged_frame['count']
    #
    # calculate frequency and score histograms
    #
    frequency_and_score_histograms(merged_frame['count'],
                                   merged_frame['score'],
                                   outdir,
                                   filestem)
    #
    # write terms
    #
    merged_frame.sort_values(by=['max_count', 'intersections'], inplace=True)
    term_filepath = os.path.join(outdir, filestem+'_terms.tsv')
    logger.debug('Writing merged terms to "%s".', term_filepath)
    merged_frame.to_csv(term_filepath, sep='\t',
                        float_format='%0.2f')
    #
    # calculate histogram of intersections
    #
    intersection_histogram(merged_frame, outdir, filestem,
                           config_obj.config_dict['plot_type'],
                           n_sets, k)


@cli.command()
@click.option('--cutoff', default=DEFAULT_SIMPLICITY_CUTOFF, help='Minimum simplicity level to unmask.')
@click.option('--plot/--no-plot', default=False, help='Plot histogram of mask fraction.')
@click.argument('infilename', type=str)
@click.argument('outfilestem', type=str)
@click.argument('setlist', nargs=-1, type=DATA_SET_VALIDATOR)
def peptide_simplicity_mask(cutoff, plot, infilename, outfilestem, setlist):
    '''Lower-case high-simplicity regions in FASTA.

    :param infilename: Name of input FASTA files for every directory in setlist.
    :param outfilestem: Stem of output filenames.
    :param cutoff: Minimum simplicity level to unmask.
    :param plot: If specified, make a histogram of masked fraction.
    :param setlist: List of defined sets to iterate over.
    :return:

    Note that this calculation is single-threaded and may be time-consuming, so
    starting multiple processes may be a good idea.
    '''
    global config_obj
    user_ctx = get_user_context_obj()
    setlist = DATA_SET_VALIDATOR.multiple_or_empty_set(setlist)
    simplicity_obj = user_ctx['simplicity_object']
    simplicity_obj.set_cutoff(cutoff)
    logger.info('Simplicity function is %s with cutoff of %d.',
                simplicity_obj.desc, cutoff)
    logger.debug('Reading from FASTA file "%s".', infilename)
    instem, ext = os.path.splitext(infilename)
    outfilename = outfilestem + ext
    logger.debug('Output FASTA file name is "%s".', outfilename)
    histfilename = outfilestem + '-hist.tsv'
    logger.debug('Output histogram file is "%s".', histfilename)
    if plot:
        plotname = outfilestem + '.' + config_obj.config_dict['plot_type']
        logger.debug('Plot to file "%s".', plotname)
    for calc_set in setlist:
        dir = config_obj.config_dict[calc_set]['dir']
        inpath = os.path.join(dir, infilename)
        outpath = os.path.join(dir, outfilename)
        shutil.copy(inpath, outpath)
        fasta  = pyfaidx.Fasta(outpath, mutable=True)
        percent_masked_list = []
        if user_ctx['first_n']:
            keys = list(fasta.keys())[:user_ctx['first_n']]
        else:
            keys = fasta.keys()
        if user_ctx['progress']:
            with click.progressbar(keys, label='%s genes processed' %calc_set,
                                   length=len(keys)) as bar:
                for key in bar:
                    masked_gene = simplicity_obj.mask(fasta[key])
                    percent_masked = 100.*num_masked(masked_gene)/len(masked_gene)
                    percent_masked_list.append(percent_masked)
        else:
            for key in keys:
                masked_gene = simplicity_obj.mask(fasta[key])
                percent_masked = 100.*num_masked(masked_gene)/len(masked_gene)
                percent_masked_list.append(percent_masked)
        fasta.close()
        #
        # histogram masked regions
        #
        (hist, bins) = np.histogram(percent_masked_list, bins=np.arange(0.,100.,100./NUM_HISTOGRAM_BINS))
        bin_centers = (bins[:-1] + bins[1:])/2.
        hist = hist*100./len(percent_masked_list)
        hist_filepath = os.path.join(dir, histfilename)
        logger.debug('writing histogram to file "%s".', hist_filepath)
        pd.Series(hist, index=bin_centers).to_csv(hist_filepath, sep='\t',
                                                  float_format='%.3f')
        #
        # plot histogram, if requested
        #
        if plot:
            plotpath = os.path.join(dir, plotname)
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(bin_centers, hist)
            plt.title('Peptide %s Simplicity Distribution with Cutoff %d'%(simplicity_obj.label.capitalize()
                                                                           ,cutoff))
            plt.xlabel('Percent of Peptide Sequence Masked')
            plt.ylabel('Percent of Peptide Sequences')
            plt.savefig(plotpath)

@cli.command()
@click.option('--force/--no-force', default=False, help='Force copy into non-empty directory.')
def install_demo_scripts(force):
    """Copy demo scripts to the working directory.

    :return:
    """
    if not force:
        files = os.listdir()
        try:
            files.remove('logs')
        except ValueError:
            pass
        if len(files) >0:
            logger.error('Current working directory is not empty.  Use --force to copy anyway.')
            sys.exit(1)
    example_files = pkg_resources.resource_listdir('aakbar', 'examples')
    for file in example_files:
        shutil.copyfile(pkg_resources.resource_filename('aakbar', os.path.join('examples', file)),
                        file)
        if file.endswith('.sh'): # set .sh files as executable
            os.chmod(file, os.stat(file).st_mode| stat.S_IEXEC)
            logger.debug('Created executable file "%s".', file)
        else:
            logger.debug('Created file "%s".', file)
    # print README
    readme = to_str(pkg_resources.resource_string('aakbar', os.path.join('examples', 'README.txt')))
    print(readme)





# -*- coding: utf-8 -*-
'''Simplicity masking and scoring classes.
'''

# external packages
import numpy as np
import pandas as pd

# module imports
from .common import *
from . import cli, get_user_context_obj, logger


class RunlengthSimplicity(SimplicityObject):
    '''Define simplicity by the number of repeated letters.

    '''
    def __init__(self, default_cutoff=DEFAULT_SIMPLICITY_CUTOFF):
        super().__init__(default_cutoff=default_cutoff)
        self.label = 'runlength'
        self.desc = 'runlength (repeated characters)'
        self.testcases += [
                 ('Simple Repeated Letters', ''),
                 ('double, beginning', 'AABCDEFGHI'),
                 (' double in middle', 'ABCDEEFGHI'),
                 ('double at end', 'ABCDEFGHII'),
                 ('double everywhere', 'AABCDDEFGG'),
                 ('triple, beginning', 'AAABCDEFGH'),
                 (' triple in middle', 'ABCDEEEFGH'),
                 ('triple at end', 'ABCDEFGIII'),
                 ('quad in middle', 'ABCDEEEEFG'),
                 ('longer string with insert',
                  'AAAAAAABCDEFGHIJKLLLLLLL'),
                 #
                 ('Mixed Patterns', ''),
                 ('mixed repeat', 'BCAABCABCA')
        ]


    def _runlength(self, s):
        return [all([s[i+j+1] == s[i] for j in range(self.cutoff-1)])
                for i in range(len(s)-self.cutoff+1)]


    def mask(self,seq):
        '''Mask high-simplicity positions in a string.

        :param s: Input string.
        :return: Input string with masked positions changed to lower-case.
        '''
        for pos in [i for i, masked in
                    enumerate(self._runlength(to_str(seq).upper()))
                    if masked]:
            if isinstance(seq, str): #strings need to have whole length set
                seq = seq[:pos] +seq[pos:pos+self.cutoff].lower() + seq[pos+self.cutoff:]
            else:
                seq[pos:pos+self.cutoff] = to_str(seq[pos:pos+self.cutoff]).lower()
        return seq



class LetterFrequencySimplicity(SimplicityObject):
   '''Define simplicity by the number of repeated letters.

   '''
   def __init__(self,
                default_cutoff=DEFAULT_SIMPLICITY_CUTOFF,
                window_size=None):
       global config_obj
       super().__init__(default_cutoff=default_cutoff)
       if window_size is None:
           try:
               self.window_size = config_obj.config_dict['letterfreq_window']
           except KeyError:
               self.window_size = DEFAULT_LETTERFREQ_WINDOW
       else:
           self.window_size = window_size
       self.label = 'letterfreq%d' %self.window_size
       self.desc = 'letter frequency in window of %d residues'%self.window_size
       self.testcases += [
                 ('Simple Repeated Letters', ''),
                 ('double, beginning', 'AABCDEFGHI'),
                 ('double in middle', 'ABCDEEFGHI'),
                 ('double at end', 'ABCDEFGHII'),
                 ('double everywhere', 'AABCDDEFGG'),
                 ('triple, beginning', 'AAABCDEFGH'),
                 ('triple in middle', 'ABCDEEEFGH'),
                 ('triple at end', 'ABCDEFGIII'),
                 ('quad in middle', 'ABCDEEEEFG'),
                 #
                 ('Pattern Repeats', ''),
                 ('doublet repeat', 'ABABABABAB'),
                 ('triplet repeat', 'ABCABCABCA'),
                 ('quad repeat', 'ABCDABCDAB'),
                 ('penta repeat', 'ABCDEABCDE'),
                 #
                 ('Mixed Patterns', ''),
                 ('mixed repeat', 'BCAABCABCA'),
                          ('long repeats with insert',
                           'SATANSPRAWNSATANSPRAWNSATANSPRAWNSAOPQTUVWXYZSATANSPAWNSATANSPAWN')

    ]


   def mask(self,seq):
        '''Mask high-simplicity positions in a string.

        :param s: Input string.
        :return: Input string with masked positions changed to lower-case.
        '''
        out_str = to_str(seq)
        end_idx = len(out_str) - 1
        byte_arr = np.array([char for char in to_bytes(out_str.upper())])
        mask_positions = set()
        #
        # test character by character for number of occurrances over window
        #
        for char in set(byte_arr): # test character by character
            char_positions = list(np.where(byte_arr == char)[0])
            while len(char_positions) >= self.cutoff:
                testpos = char_positions.pop(0)
                next_positions = char_positions[:self.cutoff-1]
                if next_positions[-1] - testpos < self.window_size:
                    mask_positions = mask_positions.union(set([testpos] + next_positions))
        #
        # mask everything
        #
        for pos in mask_positions:
            out_str = out_str[:pos] + out_str[pos].lower() + out_str[pos+1:]

        if isinstance(seq, str): # strings need to have whole length set
            seq = out_str
        else:                    # may be MutableSeq that needs lengths
            seq[:end_idx] = out_str[:end_idx]
        return seq


#
# Instantiations of classes.
#
NULL_SIMPLICITY = SimplicityObject()
RUNLENGTH_SIMPLICITY = RunlengthSimplicity()
LETTERFREQ_SIMPLICITY = LetterFrequencySimplicity()

@cli.command()
@click.option('--cutoff', default=DEFAULT_SIMPLICITY_CUTOFF, show_default=True,
              help='Maximum simplicity to keep.')
@click.option('-k', default=DEFAULT_K, show_default=True,
              help='k-mer size for score calculation.')
def demo_simplicity(cutoff, k):
    '''Demo self-provided simplicity outputs.

    :param cutoff: Simplicity value cutoff, lower is less complex.
    :param window_size: Window size for masking computation..
    :return:
    '''
    user_ctx = get_user_context_obj()
    simplicity_obj = user_ctx['simplicity_object']
    simplicity_obj.set_cutoff(cutoff)
    logger.info('Simplicity function is %s with cutoff of %d.',
                simplicity_obj.desc, cutoff)
    simplicity_obj.set_k(k)
    logger.info('           Mask window demo for %d-mers:', k)
    mask_test = 'AAAAAAAAAAaaaaAAAAAAAAAAAaaaaaAAAAAAAAAAAAaaaaaaAAAAAAAAAAAAAaaaaaaaAAAAAAAAAAAAAA'
    logger.info('      in: %s\n S-score: %s\n', mask_test,
                ''.join(['%X'%i for i in
                         simplicity_obj.score(mask_test)]))
    for desc, case in simplicity_obj.testcases:
        if case is '':
            logger.info('              %s', desc)
        else:
            masked_str = simplicity_obj.mask(case)
            logger.info('%s:\n      in: %s\n     out: %s\n S-score: %s\n',
                        desc,
                        case,
                        masked_str,
                        ''.join(['%X'%i for i in
                                 simplicity_obj.score(masked_str)
                                 ]
                                )
                        )

@cli.command()
@click.argument('window_size', default=None,
               nargs=-1)
def set_letterfreq_window(window_size):
    '''Define size of letterfreq window.
    '''
    global config_obj
    if window_size == ():
        try:
            window_size = config_obj.config_dict['letterfreq_window']
            default = ''
        except KeyError:
            window_size = DEFAULT_LETTERFREQ_WINDOW
            default = ' (default)'
        logger.info('Window size is %d residues%s',
                    window_size, default)
    elif len(window_size) > 1:
        logger.error('Only one argument for window size is permitted.')
        sys.exit(1)
    try:
        window_size = int(window_size[0])
    except ValueError:
        logger.error('Window size must be an integer value.')
        sys.exit(1)
    if window_size< 3:
        logger.error('Window size must be >=3.')
        sys.exit(1)
    config_obj.config_dict['letterfreq_window'] = window_size
    logger.info('Window size for letter-frequency simiplicity calculation is now %d residues.',
                window_size)
    config_obj.write_config_dict()


# -*- coding: utf-8 -*-
'''Global constants and common helper functions.
'''
#
# library imports
#
import logging
from datetime import datetime
from pathlib import Path # python 3.4 or later
from itertools import chain
import sys
#
# 3rd-party modules
#
import yaml
import click
import pandas as pd
#
# package imports
#
from .version import __version__
#
# global constants
#
PROGRAM_NAME = 'aakbar'
AUTHOR = 'Joel Berendzen'
EMAIL = 'joelb@ncgr.org'
COPYRIGHT = """Copyright (C) 2016, GenerisBio, LLC.  All rights reserved.
aakbar was written written under contract to  The National Center for Genome Resources.
"""
PROJECT_HOME = 'https://github.com/ncgr/aakbar'
DOCS_HOME = 'https://aakbar.readthedocs.org/en/stable'

DEFAULT_FILE_LOGLEVEL = logging.DEBUG
DEFAULT_STDERR_LOGLEVEL = logging.INFO
DEFAULT_FIRST_N = 0 # only process this many records
VERSION = __version__
STARTTIME = datetime.now()
CONFIG_FILE_ENVVAR = 'AAKBAR_CONFIG_FILE_PATH'
DEFAULT_K = 10
DEFAULT_SIMPLICITY_CUTOFF = 3
DEFAULT_LETTERFREQ_WINDOW = 10 # characters
#
# global logger object
#
logger = logging.getLogger(PROGRAM_NAME)
#
# Class definitions begin here.
#
class SimplicityObject(object):
    '''Define the interfaces needed to make simplicity calculations.

    '''
    def __init__(self, default_cutoff=DEFAULT_SIMPLICITY_CUTOFF):
        self.cutoff = default_cutoff
        self.k = None
        self.label = 'null'
        self.desc = 'no simplicity calculation'
        self.testcases = [('     non-repeated', 'ABCDEFGHIJ'),
                          ('Simple Repeated Letters', ''),
                          ('15 in a row', 'AAAAAAAAAAAAAAA')
                          ]


    def set_cutoff(self, cutoff):
        '''Set the lower bound on simplicity.

        :param cutoff:  Cutoff value, lower results in more sequence masked.
        :return: None
        '''
        if cutoff < 2:
            logger.error('Simplicity cutoff must be >=2.')
            sys.exit(1)
        else:
            self.cutoff = cutoff


    def set_k(self, k):
        '''Set the k value over which rolling scores are summed.

        :param k: Number of characters in window.
        :return:  None.
        '''
        if k < 2:
            logger.error('k must be >=2')
            sys.exit(1)
        else:
            self.k = k


    def mask(self,seq):
        '''Returns the input string.

        :param seq: Input string.
        :return: Masked copy of input string.  Must be the same length.
        '''
        return seq


    def score(self, seq):
        '''Count the number masked (by lower-case) over a window.

        :param seq: String of characters.
        :param window_size: Size of window over which to calculate.
        :return: Integer array of counts.
        '''
        is_lower = pd.Series([int(char.islower()) for char in to_str(seq)])
        return is_lower.rolling(window=self.k).sum()[self.k-1:].astype(int)

class PersistentConfigurationObject(object):
    '''Defines a persistent configuration object

    Attributes:
        :config_dict: Dictionary of configuration parameters.
        :name: Configuration filename.
        :path: Configuration filepath (absolute).
    '''
    def __init__(self, config_dir=None, name='config.yaml'):
        '''Inits the configuration dictionary

        Reads a configuration file from a subdirectory of the
        current working directory.  If that file isn't found,
        then searches in the system-specific configuration directory.
        If that file isn't found either, creates a new file in the
        directory specified by the location parameter.
        '''
        self.name = name
        self._default_dict = {'version': VERSION,
                              'simplicity_object_label': None,
                              'plot_type':'pdf',
                              'sets': [],
                              'summary': {'dir': None,
                                          'label': None},
                              }

        self._default_path = Path((click.get_app_dir(PROGRAM_NAME)+'/' + self.name))
        self._cwd_path = Path ('.' + '/.' + PROGRAM_NAME + '/' + self.name)
        if config_dir is not None:
            self.path = self._get_path_from_dir(config_dir)
        elif self._cwd_path.is_file():
            self.path = self._cwd_path
        else:
            self.path = self._default_path
        if not self.path.exists():
            self.config_dict = {}
            self.path = None
        else:
            with self.path.open('rt') as f:
                self.config_dict = yaml.safe_load(f)

    def _get_path_from_dir(self, dir):
        return Path(str(dir) + '/.' + PROGRAM_NAME +'/' + self.name).expanduser()

    def _update_config_dict(self):
        '''Update configuration dictionary if necessary
        :param config_dict: Configuration dictionary.
        :return: Updated configuration dictionary.
        '''
        try:
            if self.config_dict['version'] != VERSION:
                # Do whatever updates necessary, depending on version.
                # For now, nothing needs to be done.
                self.config_dict['version'] = VERSION
        except KeyError:
            logger.warning('Initializing config file "%s"',
                           self.path)
            self.config_dict = self._default_dict


    def write_config_dict(self, config_dict=None, dir=None):
        '''Writes a YAML configuration dictionary
        :param config_dict: Configuration dictionary
        :return: None
        '''
        if dir is None or dir is '':
            if self.path is None:
                self.path = self._default_path
        elif dir is '.':
            self.path = self._cwd_path
        else:
            self.path = self._get_path_from_dir(dir)

        if config_dict == {}:
            self.config_dict = self._default_dict
        elif config_dict is not None and config_dict != self.config_dict:
                self.config_dict = config_dict
                self._update_config_dict()

        if not self.path.parent.exists():
            # create parent directory
            logger.debug('Creating config file directory "%s"',
                          self.path.parent)
            try:
                self.path.parent.mkdir(parents=True)
            except OSError:
                logger.error('Unable to create parent directory "%s".',
                             self.path.parent)
                sys.exit(1)
        if not self.path.parent.is_dir():
            logger.error('Path "%s" exists, but is not a directory.',
                         self.path.parent)
            sys.exit(1)
        if not self.path.exists():
            logger.debug('Creating config file "%s"', self.path)
            try:
                self.path.touch()
            except OSError:
                logger.error('Path "%s" is not writable.', self.path)
                sys.exit(1)
        with self.path.open(mode='wt') as f:
            yaml.dump(self.config_dict, f)
config_obj = PersistentConfigurationObject()


class DataSetValidator(click.ParamType):
    '''Validate that set names are defined.
    '''
    global config_obj
    name = 'set'
    all_count = 0

    def convert(self, setname, param, ctx):
        '''Verify that arguments refers to a valid set.

        :param argset:
        :param param:
        :param ctx:
        :return:
        '''
        if setname == 'all':
            self.all_count += 1
            if self.all_count > 1:
                logger.error('"all" is allowed at most one time in a set list.')
                sys.exit(1)
            else:
                return tuple(config_obj.config_dict['sets'])
        elif setname not in config_obj.config_dict['sets']:
            logger.error('"%s" is not a recognized set', argset)
            sys.exit(1)
        return setname


    def multiple_or_empty_set(self, setlist):
        '''Handle special cases of empty set list or all.

        :param setlist: Setlist from validator that may be 'all' or empty.
        :return:
        '''
        # flatten any tuples due to expansion of 'all'
        if any([isinstance(setname, tuple) for setname in setlist]):
            return tuple(chain(*setlist))
        elif setlist == []:
            logger.error('Empty setlist, make sure sets are defined.')
            sys.exit(1)
        else:
            return setlist


DATA_SET_VALIDATOR = DataSetValidator()
#
# helper functions called by manyy cli functions
#
def get_user_context_obj():
    '''Returns the user context, containing logging and configuration data.

    :return: User context object (dict)
    '''
    return click.get_current_context().obj


def to_str(seq):
    '''Decode bytestring if necessary.

    :param seq: Input bytestring, string, or other sequence.
    :return: String.
    '''
    if isinstance(seq, bytes):
        value = seq.decode('utf-8')
    elif isinstance(seq, str):
        value = seq
    else:
        value = str(seq)
    return value


def to_bytes(seq):
    '''Encode or convert string if necessary.

    :param seq: Input string, bytestring, or other sequence.
    :return: Bytestring.
    '''
    if isinstance(seq, str):
        value = seq.encode('utf-8')
    elif isinstance(seq, bytes):
        value = seq
    else:
        value = bytes(seq)
    return value




# -*- coding: utf-8 -*-
'''
aakbar -- amino-acid k-mer signature tools
'''


# Developers:
# Install with
# pip install --editable .
# and execute as a module.

from setuptools import setup
from distutils.util import convert_path
import os
import sys

name = 'aakbar'
exampledir = name +'/examples/'

# restrict to python 3.4 or later
if sys.version_info < (3,4,0,'final',0):
    raise SystemExit('Python 3.4 or later is required!')

# get version from version.py
version_dict = {}
version_path = convert_path(name+'/version.py')
with open(version_path) as version_file:
    exec(version_file.read(), version_dict)
__version__ = version_dict['__version__']

# example_files is list of files in examples/ directory
exampledir = os.path.join(name, 'examples')
examples_files = []
for examplefile in ['README.txt',
                    'calculate_signatures.sh',
                    'genbank_downloader.sh',
                    'split.sh',
                    'strep10.sh']:
    examples_files.append(os.path.join(exampledir, examplefile))

setup(

    name=name,
    version=__version__,
    packages=[name],
    data_files=[('examples', examples_files)],
    url='http://github.com/ncgr/aakbar',
    keywords=['biology', 'bioinformatics', 'genomics', 'phylogenomics', 'statistics',
              'peptide', 'signatures', 'DNA', 'protein', 'sequence', 'complexity',
              'simplicity', 'alignment'],
    license='BSD',
    description='Amino-Acid k-mer Phylogenetic Signature Tools',
    long_description=open('README.rst').read(),
    author='Joel Berendzen',
    author_email='joelb@ncgr.org',
    include_package_data=True,
    zip_safe=False,
    install_requires=['biopython',
                      'click>=5.0',
                      'click_plugins',
                      'coverage',
                      'matplotlib',
                      'numpy',
                      'pandas',
                      'pyfaidx',
                      'pyyaml'],
    entry_points={
                 'console_scripts':['aakbar = aakbar:cli']
                },
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: BSD License',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Scientific/Engineering :: Bio-Informatics'
                 ]
)


# -*- coding: utf-8 -*-
'''Implements commands related to configuration parameters
'''

# module imports
from . import cli, get_user_context_obj, logger
from .common import *

# private context function
_ctx = click.get_current_context

@cli.command()
def show_config():
    '''Print location and contents of config file.

        Example:
            aakbar -v show_config
    '''
    global  config_obj

    if config_obj.config_dict == {}:
        logger.info('No configuration file was found.')
    else:
        logger.info('Configuration file path is "%s".', config_obj.path)
        for key in config_obj.config_dict.keys():
            logger.info('  %s: %s', key, config_obj.config_dict[key])


@cli.command()
@click.option('-f', '--force', is_flag=True, default=False,
              help='Overwrite existing definition.')
@click.argument('identifier', type=str)
@click.argument('dir', type=click.Path(writable=True))
def define_set(identifier, dir, force):
    '''Define an identifier and directory for a set.

    IDENTIFIER:
        a short code by which this data set will be known.
    DIR:
        absolute or relative path to files.  Must be writable.

    Example::
        aakbar define_set glymax ./Glycine_max

    '''
    global config_obj
    if identifier in config_obj.config_dict['sets'] and not force:
        logger.error('Set "%s" is already defined, use "--force" to override.',
                     identifier)
    else:
        path = dir
        if identifier in config_obj.config_dict['sets'] and force:
            logger.info('Overriding existing set "%s" definition.', identifier)
            [config_obj.config_dict['sets'].remove(identifier)
             for i in range(config_obj.config_dict['sets'].count(identifier))]
        config_obj.config_dict['sets'].append(identifier)
        config_obj.config_dict[identifier] = {'dir': str(path), 'label': identifier}
        config_obj.write_config_dict()
        logger.info('Set "%s" will refer to directory "%s".', identifier, dir)


@cli.command()
@click.argument('identifier', type=str)
@click.argument('label', type=str)
def label_set(identifier, label):
    '''Define label associated with a set.
    '''
    global config_obj
    if identifier not in config_obj.config_dict['sets']:
        logger.error('Set "%s" is not defined.', identifier)
        sys.exit(1)
    else:
        config_obj.config_dict[identifier]['label'] = label
        logger.info('Label for %s is now "%s".', identifier, label)
        config_obj.write_config_dict()


@cli.command()
@click.argument('plot_type', type=str, nargs=-1)
def set_plot_type(plot_type):
    '''Define label associated with a set.
    '''
    global config_obj
    plot_types = ['svg', 'jpg', 'ps', 'pdf', 'svgr', 'png', 'tiff', 'eps']
    if len(plot_type) == 0:
        logger.info('Supported plot types are:')
        for typename in plot_types:
            logger.info('   %s', typename)
    elif len(plot_type) > 1:
        logger.error('Only one argument for set_plot_type is allowed.')
        sys.exit(1)
    elif plot_type[0] not in plot_types:
        logger.error('Plot type "%s" is not defined.', plot_type[0])
        logger.info('Supported plot types are:')
        for typename in plot_types:
            logger.info('   %s', typename)
        sys.exit(1)
    else:
        config_obj.config_dict['plot_type'] = plot_type[0]
        logger.info('Plot type is now %s.', plot_type[0])
        config_obj.write_config_dict()


@cli.command()
@click.argument('dir', type=click.Path(writable=True))
@click.argument('label', type=str)
def define_summary(dir, label):
    '''Define summary directory and label.
    '''
    global config_obj
    path = str(Path(dir).expanduser())
    logger.info('Summary output will go in directory "%s".', path)
    config_obj.config_dict['summary']['dir'] = path
    config_obj.config_dict['summary']['label'] = label
    config_obj.write_config_dict()


@cli.command()
@click.argument('dir', type=str, default='')
def init_config_file(dir):
    '''Initialize a configuration file.

    :param dir: Optional directory in which to initialize the file.
    If not present, the system-dependent default application directory
    will be used.  If this argument is '.', then the current working
    directory will be used.  This argument accepts tilde expansions.
    '''
    global config_obj
    config_obj.write_config_dict(dir=dir, config_dict={})


@cli.command()
@click.argument('name', type=str, nargs=-1)
def set_simplicity_object(name):
    '''Select simplicity-calculation object.

    :param name: Name object of a SimplicityObject of global scope.
    :return:
    '''
    global config_obj
    known_simplicity_objects = _ctx().obj['simplicity_objects']
    if len(name) == 0:
        print('        Name       Description')
        for obj in known_simplicity_objects:
            print('%s: %s' %('{:>12}'.format(obj.label), obj.desc))
        try:
            current_simplicity_object = config_obj.config_dict['simplicity_object_label']
        except KeyError:
            current_simplicity_object = 'undefined'
        print('Current simplicity object is %s.'%current_simplicity_object)
    elif len(name) > 1:
        logger.error('Only one function may be specified')
        sys.exit(1)
    else:
        for obj in known_simplicity_objects:
            if obj.label == name[0]:
                logger.info('simplicity function is now %s.', name[0])
                config_obj.config_dict['simplicity_object_label'] = name[0]
                config_obj.write_config_dict()
                return
        logger.error('Function "%s" is not a known simplicity function.', name[0])
        sys.exit(1)

